/*

Esse programa foi desenvolvido por Mina Cintho
Arquivos de entrada: dados.txt com os atributos dos dados
		     centroides.txt com os atributos dos centróides
		     subtipos.txt com os subtipos dos dados
Arquivo de saída: clusters.txt 		     

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define infinite 99999 
#define D(x) x
#define ATRIBUTES 82
#define DATA 14393

char names[DATA][10];
char subtypes[DATA][20];



/* Print matrix */
void PrintDataMatrix (int** matrix, int rows, int columns){
        int i,j;
        for (i=0; i< rows; i++){
                D(printf("Row %d:",i+1);)
                for (j = 0; j < columns; j++){
                        D(printf("%d  ", matrix[i][j]);)
                        if (matrix[i][j]==-1) break;
                }
                D(printf("\n");)
        }
        D(printf("\n");)
}

/* Print float matrix */
void PrintFloatMatrix (float** matrix, int rows, int columns){
        int i,j;
        for (i=0; i< rows; i++){
                D(printf("Row %d:",i+1);)
                for (j = 0; j < columns; j++){
                        D(printf("%.2f ", matrix[i][j]);)
                }
                D(printf("\n");)
        }
        D(printf("\n");)
}


/* Read the data matrix */
int ** ReadingDataMatrix (int ** data){
	FILE *in;
	int i=0, num=0, k=0, j=0, dist, aux; 
        char line[1000];
	in = fopen ("dados.txt", "r");
        if (in == NULL) {
                printf("Entry file not found!\n");
                return NULL;
        }
        /* i data id, j dimension, k last 1 in data */
        while (fgets (line,1000, in)!=NULL) {
                dist=0;
                while (sscanf(line+dist,"%d%n",&num,&aux)!=EOF){                 
                        if (num ==1) data[i][k++]=j;
                        else if (num !=0) D(printf("Reading not valid data: %d \n",num);)
			j++;
                        dist = dist+aux;
                }
                if (j>ATRIBUTES) D(printf("Reading too many atributes\n");)
		if (k<ATRIBUTES) data[i][k]=-1;
                j=0;
                i++;
		k=0;
        }
        fclose (in);	
	return data;
}


/* Read the centroid matrix */
float ** ReadingCentroids (float ** centroids){
	FILE *in;
	int i=0, num=0, j=0, dist, aux; 
        char line[1000];
	in = fopen ("centroides.txt", "r");
        if (in == NULL) {
                printf("Centroid file not found!\n");
                return NULL;
        }
        while (fgets (line,1000, in)!=NULL) {
                dist=0;
                while (sscanf(line+dist,"%d%n",&num,&aux)!=EOF){
                        if (num == 1)  centroids[i][j++]=1;
                        else if (num==0)  centroids[i][j++]=0;
                        else D(printf("Reading not valid data: %d \n",num);)
                        dist = dist+aux;
                }
                if (j>ATRIBUTES) D(printf("Reading too many atributes\n"));
                i++;
		j=0;
        }
        fclose (in);	
	return  centroids;
}



/* Alloc floatting matrix */
float ** AllocMatrixfloat (float ** matrix, int rows, int columns){
	int i;
	matrix = malloc( rows *(sizeof (float*)));
	for (i=0; i< rows;i++)
		matrix[i] = malloc(columns * (sizeof (float)));
	return matrix;
}


/* Alloc int matrix */
int ** AllocMatrix (int ** matrix, int rows, int columns){
	int i;
	matrix = malloc( rows *(sizeof (int*)));
	for (i=0; i< rows;i++)
		matrix[i] = malloc(columns * (sizeof (int)));
	return matrix;
}

/* Alloc vector */
int * AllocVector (int * vector, int size){
	vector = malloc( size *(sizeof (int)));
	return vector;
}

/* Init vector */
int * InitVector (int * vector, int size, int value){
        int i;
        for (i=0;i<size;i++){
                vector[i]=value;
        }
	return vector;
}

/* Init Matrix */
float ** InitMatrix (float ** matrix, int rows, int columns, float value){
	int i,j;
	for (i=0; i< rows;i++)
	        for (j=0; j< columns;j++)
		        matrix[i][j]=value;
	return matrix;
}

/* Calculate the distance of the element dat from data to the centroid cent  */
int CalculateDistance (int ** data, float ** centroids, int dat, int cent ){
	int dist=0, i=0,k=0;
	/* k index of centroids, i index of data, j last index of data with 1 */
	/* goes through the data element */
        while (i<ATRIBUTES && data[dat][i]!=-1) {
                /* goes through the centroid element while it is smaller than the elements 0 in data */
                while (k < data[dat][i]){
                        /* If the centroid has 1 in position k it rises the distance */
                       if (centroids[cent][k]==1){
                                dist++;
                       }
                       k++;
                }
                /* If the position k in the centroid is 0 it raises the distance */                
                if (centroids[cent][k]==0) dist++;
                k++;
                i++;
        } 
        /* if it haven't gone through all the lelements of the centroid */
        while (k<ATRIBUTES){
               if (centroids[cent][k]==1){
                        dist++;
               }
               k++;
        }       
	return dist;
}


int ChooseClusters(int ** data, float ** centroids,int * weights, int * clusterSize,  float ** mean, int numClusters){
        int i=0,j=0,k=0, min=infinite, aux=0, w= -1, changes=0;
        /* For each data */
	for (i=0; i< DATA; i++){
                /* For each centroid */
                for (j=0;j<numClusters;j++){
                        /* Calculates the distance of the element to the centroid */
                        aux = CalculateDistance (data, centroids, i,j);
                        /* If the distance to the centroid is the smallest found  */
		        if (aux < min) { 
                                min=aux;
                                w=j;
                        }
                }
                /* If the element is in another cluster */
                if (weights[i]!=w){
                        changes++;                                
                        weights[i]=w;
                }
                k=0;
                /* Put the element in mean */
                while(k<ATRIBUTES && data[i][k]!=-1) {
                        mean[w][data[i][k]] += 1;
                        k++;
                }
                clusterSize[w]++;
                min = infinite;
                w=-1;   	
        }
        return changes;
}

/* Calculate the new centroids */
float ** newCentroids (float ** centroids, float ** means, int * clusterSize, int numClusters){
        int i=0, j=0;
        for (i=0; i<numClusters; i++){
                for (j=0;j<ATRIBUTES;j++){
                       if (clusterSize[i]>0) {
                       	centroids[i][j] = means[i][j]/clusterSize[i];
                       }
                       else centroids[i][j] = -10000;
                }
        }
        return centroids;
}


void Kmeans (int ** data, float ** centroids, int * weights, float ** mean,int * clusterSize, int numClusters ){
	int changes=1, count=0;	
	while (changes > 0){
                /* While clusters are still changing */
		changes=0;		
		changes = ChooseClusters(data,centroids,weights,clusterSize,mean,numClusters);
		centroids = newCentroids (centroids, mean, clusterSize, numClusters);
		count++;
	}

}




int main (){
	int  numClusters=0;
	int ** data=NULL,  * weights=NULL,  * clusterSize=NULL;
        float ** centroids = NULL, **mean =NULL;

	FILE *in;
	int i=0, dist,aux,temp; 	
	char  name[40];	
        char line[1000];     
        int j,k, l,m;
        FILE * out;
        out = fopen("clusters.txt", "w");
        D(fprintf (out, "Clusters:\n");)        
	
	in = fopen ("subtipos.txt", "r");	
        if (in == NULL) {
                printf("Subtype file not found!\n");

        }

	printf("Number of clusters : ");
	scanf("%d", &numClusters);
        

        while (fgets (line,1000, in)!=NULL) {
                dist=0;
                temp=0;
                while (sscanf(line+dist,"%s%n",name,&aux)!=EOF){
	
			if (temp%2==0) strcpy(names[i],name);
			else strcpy(subtypes[i],name);
                        dist = dist+aux;
                        temp++;			
                }
                i++;
                

       	}    
        fclose (in);	



        
	data = AllocMatrix (data, DATA, ATRIBUTES);
	centroids = AllocMatrixfloat(centroids, numClusters, ATRIBUTES);

	weights = AllocVector(weights, DATA);

        clusterSize= AllocVector(clusterSize, numClusters);
        mean = AllocMatrixfloat(mean, numClusters, ATRIBUTES);
        
        clusterSize= InitVector(clusterSize, numClusters,0);
        weights = InitVector(weights, DATA, -1);
        mean = InitMatrix (mean, numClusters, ATRIBUTES, 0);

	data = ReadingDataMatrix (data);
        centroids = ReadingCentroids (centroids);
        

 	Kmeans(data, centroids, weights, mean,clusterSize, numClusters);


        for (i=0; i<numClusters ;i++){
                D(fprintf(out, "\n\n Cluster %d \n",i+1);)
                for (j=0; j< DATA; j++){
                        if (weights[j]==i){
                        	 D(fprintf(out, "element ");)
                        	for (m=0;m<11;m++){
                        		 D(fprintf(out, "%c",names[j][m]);)
                        	}
                                D(fprintf(out, " subtype %s \n", subtypes[j]);)
                                k=0;
                                l=0;
                                fprintf(out,"\n");
                        }
                }
        }
        fclose(out);	

	return 0;
}
