/*

Esse programa foi desenvolvido por Mina Cintho
Arquivos de entrada: dados.txt com os atributos dos dados     
Arquivo de saída: matlab.m (entrada para plotar dendrograma no matlab) 		     

*/


#include <stdio.h>
#include <stdlib.h>
#define T(y) 
#define D(x) x
#define DATA 14393
#define ATRIBUTES 82
#define NUMEDGES ((DATA*DATA/2)-(DATA/2))

/* Vector that sets the color of the vertice */
static int color[DATA];
/* Vector that saves the size of each color group */
static int size[DATA];
/* Vector that saves the order of the union of the clusters */
static int clusters [DATA-1][3];
/* Matrix with the data */
int data[DATA][ATRIBUTES];
/* Matrix with the index of the positions that had diferences */
int diferences [DATA-1] [ATRIBUTES+3];


typedef struct { 
   int v, w; 
   int dist; 
} Edge;

/* Vector with all of the edges in the graph */
Edge edges[NUMEDGES];

/* Insert an arc */
Edge InsertEdge (int v, int w, int d) {
   Edge edge;
   edge.v = v; edge.w = w;
   edge.dist = d;
   return edge;
}


struct g {
	int Edges; 
	int Vertex;
	int adj[DATA][DATA]; 
};

typedef struct g* Graph;



/* Set the edges and it's costs with -1 */
void SetEdgesVector () {
        int i;
        for (i=0; i< NUMEDGES; i++){
                edges[i].v = edges[i].w = edges[i].dist = -1;
        }
}


/* Print all edges */
void PrintEdges (Edge edges[], int size){
        int i;
        for (i=0; i< size; i++)        
                if (edges[i].dist!=-1) D(printf("Edge from %d to %d with distance %d \n", (edges[i].v)+1, (edges[i].w)+1, edges[i].dist);)
        D(printf("\n");)
}

/* Print some matrix */
void PrintDataMatrix (int rows, int columns){
        int i,j;
        for (i=0; i< rows; i++){
                D(printf("Row %d:",i+1);)
                for (j = 0; j < columns; j++){
                        D(printf("%d", data[i][j]);)
                }
                D(printf("\n");)
        }
        D(printf("\n");)
}



/* Set the graph */
Graph SetGraph (Graph G, int vert) {	
	G = malloc(sizeof *G);
	G->Vertex = vert;
	G->Edges = (vert*vert);
	return G;
}

/* Insert an edge into the graph */
void InsertInGraph (Graph G, int v, int w, int distance) {
		G->adj[v][w] = distance;
		G->adj[w][v] = distance;
}

/* Calculates the distances of the vertice */
void SetEuclidian(Graph G, int atribute){
        int i,j,k,sum=0;
        for (i=0; i< G->Vertex; i++)
                for (j=0;j< i;j++) {
                        for (k=0; k < atribute;k++){
                                sum = sum + ((data[i][k]-data[j][k]) * (data[i][k]-data[j][k]));
                        }
                        InsertInGraph(G,i,j,sum);
                        sum=0;
                }        
}


/*  Set each vertex with it's on color and it's group size with 1 */
void SetColorAndSize(int n) { 
        int i;
        for (i = 0; i < n; i++) { 
                color[i] = i; 
                size[i] = 1; 
        }
}


/* Check if the vertex are in the same union tree, if so return 1*/
static int find(int v) {  
        while (v != color[v]) 
                v = color[v]; 
        return v; 
}

/* Return 1 if the vertice have the same colors */
int UFfind(int v, int w) { 
        return (find(v) == find(w)); 
}

/* Paints two different groups with the same color */
void UFunion(int v, int w) { 
        if (v == w) return;
        if (size[v] < size[w]) { 
                color[v] = w; 
                size[w] += size[v]; 
        }
        else { 
                color[w] = v; 
                size[v] += size[w]; 
        }
}

/* Compares the costs of two edges */
int CompareDist (const void *a, const void *b){
        int iA, iB;
        iA = ((Edge *)a)->dist; 
        iB = ((Edge *)b)->dist;
    return iA  - iB;
}


/* Insert the edges of the graph into the edges vector */
void EdgesList(Graph G){
        int i,j,k=0;
        for (i=0; i< G->Vertex; i++){
                for (j=0; j < i;j++){
                        edges[k++] = InsertEdge (i, j, G->adj[i][j]);         
                }     
        }
}

void PrintOutputToMatlab(){
        int i;
        FILE *out;
        out = fopen ("matlab.m", "w");
        fprintf(out, "Clusters = [");
        for (i=0; i<DATA-2;i++){
	        fprintf(out, "%d %d %d \n",clusters [i][0]+1,clusters [i][1]+1,clusters [i][2]);
        }
        fprintf(out, "%d %d %d ]", clusters [DATA-2][0]+1,clusters [DATA-2][1]+1,clusters [DATA-2][2]);
        fclose (out);
        
}

void Kruskal (Graph G){ 
        int i, k;
        SetEdgesVector (); 
        /* Insert the edges of the graph into the vector edges */
        EdgesList(G);
        /* Sort the edges by the distance */
        qsort(edges,NUMEDGES, sizeof(Edge), CompareDist);
        /* Sets each element with it's own color and the vector size */
        SetColorAndSize(G->Vertex);
        /* Search the vector with the edges */
        for (i = k = 0; i < NUMEDGES && k < G->Vertex-1; i++)
                /* Verify if the vertice with the smaller distance already have nodes with the same color */
                if (!UFfind(edges[i].v, edges[i].w)) {
                        /* If they do not have the same color it paints them with the same color */
                        UFunion(edges[i].v, edges[i].w);
                        /* Register the union of the elements */
                        clusters[k][0]=edges[i].v;
                        clusters[k][1]=edges[i].w;
                        clusters[k][2]=edges[i].dist;
                        k++;
                }
        PrintOutputToMatlab();
}


/* Print some diagonal matrix */
void PrintDiagonalMatrix (Graph G){
        int i,j;
        for (i=0; i< G->Vertex; i++){
                D(printf("Row %d:",i+1);)
                for (j = 0; j < i; j++){
                        D(printf("%3d ", G->adj[i][j]);)
                }
                D(printf("\n");)
        }
        D(printf("\n");)
}


/* Print some diagonal matrix */
void PrintMatrix ( int rows, int columns){
        int i,j;
        for (i=0; i< rows; i++){
                j=0;
                while (j < columns && diferences[i][j]!=-1 ){
                        D(printf("%d ", diferences[i][j]);)
                        j++;
                }
                D(printf("\n");)
        }
        D(printf("\n");)
}




/* Put in the matrix diferences the elements and the distance of the vertex used and the positions with diferent values */
void ComputeDiferences (){
        int i,j,k,sum;
        for (i=0; i<DATA-1; i++){
                diferences[i][0]=clusters[i][0]+1;
                diferences[i][1]=clusters[i][1]+1;
                diferences[i][2]=clusters[i][2];
                j=3;
                for (k=0; k < ATRIBUTES;k++){
                        sum = ((data[clusters[i][0]][k]-data[clusters[i][1]][k]) * (data[clusters[i][0]][k]-data[clusters[i][1]][k]));
                        if (sum==1)
                               diferences[i][j++]=k;
                }
                 diferences[i][j]=-1;
        }
}



/* Calculates the distances of the vertice */
void SetEuclidianDiferences(){
        int i,j,k=2, out=0;
        for (i=0; i< DATA-1; i++)
                for (j=0;j< i;j++) {
                        while (k<ATRIBUTES-1 && diferences[i][k]!=-1 && diferences[j][k]!=-1 && out==0){
                                if (diferences[i][k]!=diferences[j][k] )out=1;
                                k++;
                        }

                        out=0;
                        k=3;
                }                      
}



int main (){
	int dataSize = DATA, atributeSize = ATRIBUTES,i=0,j=0, num,dist,aux;
	Graph G = NULL;
        /* Archive with the data */
        FILE *in;
        char line[1000];
        /*Gets the data from text*/
        in = fopen ("dados.txt", "r");
        if(in == NULL) {
		printf("Arquivo de entrada não encontrado!\n");
		return 1;
	}
        while (fgets (line,1000, in)!=NULL) {
                dist=0;
                while (sscanf(line+dist,"%d%n",&num,&aux)!=EOF){
                        if (j>ATRIBUTES) D(printf("Lendo mais de %d atributos\n", ATRIBUTES);)
                        if (num == 0) data[i][j++]=0; 
                        else if (num ==1) data[i][j++]=1;
                        else D(printf("Leitura de valor diferente de um e zero: %d \n",num);)
                        dist = dist+aux;
                }
                j=0;
                i++;
        }
        fclose (in);

        /* Set graph and adjacency matrix */
	G = SetGraph (G, dataSize);
        SetEuclidian(G, atributeSize);

        /* Kruskal */    
        Kruskal (G);
        
        /* Calculate the distances of the vertices */
        ComputeDiferences ();
        D(printf("\n\n");)
        D(printf("Diferences: \n");)
        PrintMatrix(DATA-1, ATRIBUTES+2);
        
        /* Statistics of the positions used */
        SetEuclidianDiferences();



        
	return 0;
}
